class SeqOp:
    pass


class SeqLoss(SeqOp):
    def __init__(self):
        self.time = 0
        self.mem = 0

    def __str__(self):
        return "Loss"


# ==========================


# ==========================
# ====== Seq Block Op ======
# ==========================
# -> Attributes : .body(SeqAtomOp list) ; .name ; .var ; .mem/time
# -> Methods    : __str__
# -> Subclasses : Fn,Fc,Fe,Bwd
class SeqBlockOp(SeqOp):
    def __init__(self, name, index, option):
        self.name = name
        self.index = index
        self.option = option

    def __str__(self):
        if self.option is None:
            return f"{self.name}_{self.index}"
        return f"{self.name}_{self.index}_{self.option}"

class SeqBlockFn(SeqBlockOp):
    def __init__(self, *args):
        super().__init__("Fn", *args, None)


class SeqBlockFc(SeqBlockOp):
    def __init__(self, *args):
        super().__init__("Fc", *args, None)


class SeqBlockFe(SeqBlockOp):
    def __init__(self, *args):
        super().__init__("Fe", *args)


class SeqBlockBwd(SeqBlockOp):
    def __init__(self, *args):
        super().__init__("Bwd", *args)

class RK_Sequence:
    def __init__(self, l=None):
        if l:
            self.seq = l
        else:
            self.seq = []

    def __str__(self):
        return ", ".join([str(o) for o in self.seq])

    def insert(self, op: SeqOp):
        self.seq.append(op)

    def insert_seq(self, sub_seq):
        self.seq.extend(sub_seq.seq)



# import logging
# import math
from typing import Dict, Any
import numpy as np
from rockmate.def_op import RunOp, DelOp, OpSchedule
from mip import Model, xsum, maximize, BINARY, minimize, OptimizationStatus
from rkgb.utils.global_vars import solver_name


class ModelMIP:
    """
    The MIP model will build the ILP model by given Kgraph and budget.
    """

    def __init__(
        self,
        kg,
        budget: int,
        save_budget: int,
        gcd=None,
        gurobi_params: Dict[str, Any] = {},
    ):
        self.kg = kg
        self.time = [kcn.time for kcn in self.kg.list_kcn]
        self.gcd = gcd if gcd else 1
        self.budget = budget / self.gcd
        self.save_budget = save_budget / self.gcd
        self.overhead = [kcn.overhead.v / self.gcd for kcn in self.kg.list_kcn]
        self.mem = [kdn.mem.v / self.gcd for kdn in self.kg.list_kdn]
        self.feasible = None
        self.solve_time = None

        self.output_indices = [
            self.kg.list_kdn.index(n) for n in [self.kg.output_kdn_grad]
        ]
        #  self.kg.output_kdn_data]]
        self.protected_indices = []
        self.loss_idx = self.kg.list_kcn.index(self.kg.loss_kcn)
        T = len(self.kg.list_kcn)
        I = len(self.kg.list_kdn)

        self.md = Model(
            f"rockmateMILP_{T}_{budget}", solver_name=solver_name[0]
        )
        if gurobi_params is not None:
            for k, v in gurobi_params.items():
                setattr(self.md.Params, k, v)

        _deps_d = [
            [self.kg.list_kcn.index(kcn) for kcn in self.kg.list_kdn[i].deps]
            for i in range(I)
        ]
        # _deps_c = [[self.kg.list_kdn.index(kdn)
        #             for kdn in self.kg.list_kcn[i].deps_real] for i in range(T)]
        _users_d = [
            [
                self.kg.list_kcn.index(kcn)
                for kcn in self.kg.list_kdn[i].users_real
                if kcn in self.kg.list_kcn
            ]
            for i in range(I)
        ]
        # TODO: there's user in the next graph?
        # return [self.kg.list_kcn.index(kcn)
        #         for kcn in self.kg.list_kdn[i].users_real]
        _users_c = [
            [self.kg.list_kdn.index(kdn) for kdn in self.kg.list_kcn[i].users]
            for i in range(T)
        ]

        self.create_list = [
            (k, i)
            for k, kcn in enumerate(self.kg.list_kcn)
            for i in _users_c[k]
        ]
        self.delete_list = [
            (k, i)
            for i, kdn in enumerate(self.kg.list_kdn)
            for k in _deps_d[i] + _users_d[i]
        ]

        Cr = len(self.create_list)
        De = len(self.delete_list)
        # ======build varaibles======
        self.R = self.md.add_var_tensor((T, T), "R", var_type=BINARY)
        self.S = self.md.add_var_tensor((T, Cr), "S", var_type=BINARY)
        self.P = self.md.add_var_tensor((T, I), "P", var_type=BINARY)
        self.create = self.md.add_var_tensor((T, Cr), "create", var_type=BINARY)
        self.delete = self.md.add_var_tensor((T, De), "delete", var_type=BINARY)

        # define objective function
        self.md.objective = minimize(
            xsum(
                self.R[t, i] * self.time[i] for t in range(T) for i in range(T)
            )
        )

        # ======build constraints======
        self.md.add_constr(
            xsum(self.R[t, i] for t in range(T) for i in range(t + 1, T)) == 0,
        )
        self.md.add_constr(
            xsum(
                self.S[t, j]
                for j in range(Cr)
                for t in range(self.create_list[j][0] + 1)
            )
            == 0,
        )
        self.md.add_constr(
            xsum(
                self.P[t, i]
                for i in range(I)
                for t in range(min(_deps_d[i]) + 1)
            )
            == 0,
        )
        self.md.add_constr(xsum(self.R[t, t] for t in range(T)) == T)
        self.md.add_constr(
            xsum(self.R[t, self.loss_idx] for t in range(T)) == 1
        )  # fwd_loss can only run once

        for t in range(T):
            for j in range(Cr):
                self.md.add_constr(
                    self.S[t, j] <= self.P[t, self.create_list[j][1]],
                )
        for t in range(T - 1):
            for i in range(Cr):
                self.md.add_constr(
                    self.S[t + 1, i]
                    <= self.S[t, i] + self.R[t, self.create_list[i][0]],
                )
        # ensure all computations are possible
        for t in range(T):
            for j, (k, i) in enumerate(self.create_list):
                for k_ in _users_d[i]:
                    self.md.add_constr(
                        self.R[t, k_] <= self.R[t, k] + self.S[t, j],
                    )

        self.alive = {}
        for t in range(T):
            for eidx, (k, i) in enumerate(self.delete_list):
                self.alive[(t, k, i)] = self.P[t, i]
                self.alive[(t, k, i)] += xsum(
                    self.create[t, eidx_c]
                    for eidx_c, (k_, i_) in enumerate(self.create_list)
                    if i_ == i and k_ <= k
                )
                self.alive[(t, k, i)] -= xsum(
                    self.delete[t, eidx_d]
                    for eidx_d, (k_, i_) in enumerate(self.delete_list)
                    if i_ == i and k_ <= k
                )
                self.md.add_constr(self.alive[(t, k, i)] >= 0)
                self.md.add_constr(self.alive[(t, k, i)] <= 1)
                if (k, i) in self.create_list:
                    didx = self.delete_list.index((k, i))
                    self.md.add_constr(
                        self.alive[(t, k, i)] + self.delete[t, didx]
                        >= self.R[t, k],
                    )

            for eidx, (k, i) in enumerate(self.create_list):
                self.md.add_constr(self.create[t, eidx] <= self.R[t, k])
            for i in range(I):
                if t + 1 < T:
                    self.md.add_constr(
                        self.P[t + 1, i]
                        == self.alive[(t, max(_deps_d[i] + _users_d[i]), i)],
                    )
                else:  # if i not in self.output_indices:
                    # in the end of bwd, del everything
                    self.md.add_constr(
                        self.alive[(t, max(_deps_d[i] + _users_d[i]), i)] == 0,
                    )

        def _num_hazards(t, i, k):
            if i in self.protected_indices:
                return _max_num_hazards(t, i, k)
            if t + 1 < T:
                return (
                    1
                    - self.R[t, k]
                    + self.P[t + 1, i]
                    + xsum(self.R[t, j] for j in _users_d[i] if j > k)
                )
            return (
                1
                - self.R[t, k]
                + xsum(self.R[t, j] for j in _users_d[i] if j > k)
            )

        def _max_num_hazards(t, i, k):
            num_uses_after_k = sum(1 for j in _users_d[i] if j > k)
            if t + 1 < T:
                return 2 + num_uses_after_k
            return 1 + num_uses_after_k

        # delete when not needed
        # for t in range(T):
        #     for eidx, (k, i) in enumerate(self.delete_list):
        #         self.md.add_constr(1 - self.delete[t, eidx]
        #                              <=
        #                             _num_hazards(t, i, k))

        # don't delete if still needed
        for t in range(T):
            for eidx, (k, i) in enumerate(self.delete_list):
                self.md.add_constr(
                    _max_num_hazards(t, i, k) * (1 - self.delete[t, eidx])
                    >= _num_hazards(t, i, k),
                )

        self.U = {}
        for t in range(T):
            self.U[(t, 0)] = (
                xsum(self.P[t, i] * self.mem[i] for i in range(I))
                + xsum(
                    self.create[t, eidx] * self.mem[i]
                    for eidx, (k_, i) in enumerate(self.create_list)
                    if k_ == 0
                )
                + xsum(
                    self.delete[t, eidx] * self.mem[i]
                    for eidx, (k_, i) in enumerate(self.delete_list)
                    if k_ == 0
                )
            )

        for t in range(T):
            for k in range(1, T):
                self.U[(t, k)] = (
                    self.U[(t, k - 1)]
                    + xsum(
                        self.create[t, eidx] * self.mem[i]
                        for eidx, (k_, i) in enumerate(self.create_list)
                        if k_ == k
                    )
                    - xsum(
                        self.delete[t, eidx] * self.mem[i]
                        for eidx, (k_, i) in enumerate(self.delete_list)
                        if k_ == k
                    )
                )
        for t in range(T):
            for k in range(T):
                self.md.add_constr(self.U[(t, k)] >= 0)
                self.md.add_constr(
                    self.U[(t, k)]
                    + self.R[t, k] * self.overhead[k]
                    + xsum(
                        self.mem[i_] * self.delete[t, eidx_d]
                        for eidx_d, (k_, i_) in enumerate(self.delete_list)
                        if k == k_
                    )
                    <= self.budget,
                )
                if t == T // 2 and self.save_budget:
                    self.md.add_constr(self.U[(t, k)] <= self.save_budget)

    def add_abar_constraint(self, save_budget):
        T = len(self.kg.list_kcn)
        self.save_budget = save_budget / self.gcd
        for k in range(T):
            t = T // 2
            self.md.add_constr(self.U[(t, k)] <= self.save_budget)

    def solve(self):

        # self.md.message("\n\nRestarting solve\n\n")
        self.md.optimize()

        infeasible = self.md.status == OptimizationStatus.INFEASIBLE
        if infeasible:
            self.feasible = False
        else:
            # if self.md.solCount < 1:
            #     raise ValueError(
            #         "Model status is {}, but solCount is {}".format(
            #             self.md.status, self.md.solCount
            #         )
            #     )
            self.feasible = True

    def schedule(self, kg=None):
        kg = kg if kg else self.kg
        assert self.feasible, "Cannot schedule an infeasible model!"
        T = len(kg.list_kcn)
        I = len(kg.list_kdn)

        op_list = []
        alive_list = []
        alive_status = np.zeros(I + 2, dtype=bool)
        alive_status[-1] = 1  # input_data_kdn
        for t in range(T):
            for k in range(T):
                if self.R[t, k] == 1:
                    kcn = kg.list_kcn[k]
                    if "loss" in kcn.name:
                        op_list.append(RunOp(kcn))
                        alive_list.append(alive_status.copy())
                        alive_status[kg.list_kdn.index(kg.output_kdn_grad)] = 1
                        # alive_status[kg.list_kdn.index(kg.output_kdn_data)] = 0
                    for eidx, (k_, i) in enumerate(self.create_list):
                        if k == k_ and self.create[t, eidx] == 1:
                            alive_status[i] = 1
                    op_list.append(RunOp(kcn))
                    alive_list.append(alive_status.copy())
                    # for i in range(I):
                    #     if self.alive[(t,k,i)].getValue():
                    #         alive_list[-1][i] = 1
                for eidx, (k_, i) in enumerate(self.delete_list):
                    if k == k_ and self.delete[t, eidx] == 1:
                        kdn = kg.list_kdn[i]
                        if "phantom" in kdn.name:
                            alive_status[i] = 0
                            op_list.append(DelOp(kdn))
                            alive_list.append(alive_status.copy())
                for eidx, (k_, i) in enumerate(self.delete_list):
                    if k == k_ and self.delete[t, eidx] == 1:
                        kdn = kg.list_kdn[i]
                        if "phantom" not in kdn.name:
                            alive_status[i] = 0
                            op_list.append(DelOp(kdn))
                            alive_list.append(alive_status.copy())
        for i, op in enumerate(op_list):
            if "loss" in op.name:
                loss_i = i
                break

        input_kdn = kg.input_kdn_data
        # if "src" not in input_kdn.name:
        #     del_input_op = DelOp(input_kdn)
        #     del_input_idx = len(op_list)
        #     for i, op in enumerate(op_list):
        #         if isinstance(op, RunOp) and input_kdn in op.deps_global:
        #             del_input_idx = i + 1
        #     op_list.insert(del_input_idx, del_input_op)
        #     alive_status = alive_list[del_input_idx - 1]
        #     alive_status[-1] = 0
        #     alive_list.insert(del_input_idx, alive_status)

        fwd_sched = OpSchedule(
            op_list[: loss_i + 1],
            alive_list[: loss_i + 1],
            kg.input_kdn_data,
            kg.input_kdn_grad,
            kg.output_kdn_data,
            kg.list_kdn,
        )
        bwd_sched = OpSchedule(
            op_list[loss_i + 1 :],
            alive_list[loss_i + 1 :],
            kg.input_kdn_data,
            kg.input_kdn_grad,
            kg.output_kdn_data,
            kg.list_kdn,
        )
        # fwd_sched.del_input(kg)
        return fwd_sched, bwd_sched



# ==========================
# modified version of rotor algo
# contains RK_Sequence builder -> depends on RK_Chain
# based on rotor/algorithms/persistent.py
# ==========================

from rkgb.utils import print_debug
from rockmate.def_chain import RK_Chain
from rockmate.def_sequence import (
    SeqBlockFn,
    SeqBlockFc,
    SeqBlockFe,
    SeqBlockBwd,
    SeqLoss,
    RK_Sequence,
)

# ==========================
# ==== DYNAMIC PROGRAM =====
# ==========================


def psolve_dp_functionnal(chain, mmax, opt_table=None):
    """Returns the optimal table:
    Opt[m][lmin][lmax] : int matrix
        with lmin = 0...chain.length
        and  lmax = lmin...chain.length (lmax is not included)
        and  m    = 0...mmax
    What[m][lmin][lmax] is :
        (True, k) if the optimal choice is a chain chkpt
        -> ie F_e this block with solution k
        (False,j) if the optimal choice is a leaf  chkpt
        -> ie F_c and then F_n (j-1) blocks
    """

    ln = chain.ln
    fw = chain.fw
    bw = chain.bw
    cw = chain.cw
    cbw = chain.cbw
    fwd_tmp = chain.fwd_tmp
    bwd_tmp = chain.bwd_tmp
    ff_fwd_tmp = chain.ff_fwd_tmp
    ff_fw = chain.ff_fw
    nb_sol = chain.nb_sol

    opt, what = opt_table if opt_table is not None else ({}, {})
    # opt = dict()
    # what = dict()

    def opt_add(m, a, b, time):
        if not m in opt:
            opt[m] = dict()
        if not a in opt[m]:
            opt[m][a] = dict()
        opt[m][a][b] = time

    def what_add(m, a, b, time):
        if not m in what:
            what[m] = dict()
        if not a in what[m]:
            what[m][a] = dict()
        what[m][a][b] = time

    # -> Last one is a dict because its indices go from i to l.
    # -> Renumbering will wait for C implementation

    # -- Initialize borders of the tables for lmax-lmin = 0 --
    def case_d_0(m, i):
        possibilities = []
        for k in range(nb_sol[i]):
            limit = max(
                cw[i] + cbw[i + 1][k] + fwd_tmp[i][k],
                cw[i] + cbw[i + 1][k] + bwd_tmp[i][k],
            )
            if m >= limit:
                possibilities.append((k, fw[i][k] + bw[i][k]))
        if possibilities == []:
            opt_add(m, i, i, float("inf"))
        else:
            best_sol = min(possibilities, key=lambda t: t[1])
            opt_add(m, i, i, best_sol[1])
            what_add(m, i, i, (True, best_sol[0]))
        return opt[m][i][i]

    # -- dynamic program --
    nb_call = 0

    def solve_aux(m, a, b):
        if (m not in opt) or (a not in opt[m]) or (b not in opt[m][a]):
            nonlocal nb_call
            nb_call += 1
            if a == b:
                return case_d_0(m, a)
            #  lmin = a ; lmax = b
            mmin = cw[b + 1] + cw[a + 1] + ff_fwd_tmp[a]
            if b > a + 1:
                mmin = max(
                    mmin,
                    cw[b + 1]
                    + max(
                        cw[j] + cw[j + 1] + ff_fwd_tmp[j]
                        for j in range(a + 1, b)
                    ),
                )
            if m < mmin:
                opt_add(m, a, b, float("inf"))
            else:
                #  -- Solution 1 --
                sols_later = [
                    (
                        j,
                        (
                            sum(ff_fw[a:j])
                            + solve_aux(m - cw[j], j, b)
                            + solve_aux(m, a, j - 1)
                        ),
                    )
                    for j in range(a + 1, b + 1)
                    if m >= cw[j]
                ]
                if sols_later:
                    best_later = min(sols_later, key=lambda t: t[1])
                else:
                    best_later = None

                #  -- Solution 2 --
                # -> we can no longer use opt[i][i] because the cbw
                # -> now depend on the F_e chosen.
                sols_now = []
                for k in range(nb_sol[a]):
                    mem_f = cw[a + 1] + cbw[a + 1][k] + fwd_tmp[a][k]
                    mem_b = cw[a] + cbw[a + 1][k] + bwd_tmp[a][k]
                    limit = max(mem_f, mem_b)
                    if m >= limit:
                        time = fw[a][k] + bw[a][k]
                        time += solve_aux(m - cbw[a + 1][k], a + 1, b)
                        sols_now.append((k, time))
                if sols_now:
                    best_now = min(sols_now, key=lambda t: t[1])
                else:
                    best_now = None

                # -- best of 1 and 2 --
                if best_later is None and best_now is None:
                    opt_add(m, a, b, float("inf"))
                elif best_later is None or (
                    best_now is not None and best_now[1] < best_later[1]
                ):
                    opt_add(m, a, b, best_now[1])
                    what_add(m, a, b, (True, best_now[0]))
                else:
                    opt_add(m, a, b, best_later[1])
                    what_add(m, a, b, (False, best_later[0]))
        return opt[m][a][b]

    solve_aux(mmax, 0, ln)

    print_debug(f"Nb calls : {nb_call}")
    return (opt, what)


# ==========================
#  ==== SEQUENCE BUILDER ====
# ==========================


def pseq_builder(chain, memory_limit, opt_table):
    # returns :
    # - the optimal sequence of computation using mem-persistent algo
    mmax = memory_limit - chain.cw[0]
    # opt, what = solve_dp_functionnal(chain, mmax, *opt_table)
    opt, what = opt_table
    #  ~~~~~~~~~~~~~~~~~~
    def seq_builder_rec(lmin, lmax, cmem):
        seq = RK_Sequence()
        if lmin > lmax:
            return seq
        if cmem <= 0:
            raise ValueError(
                "Can't find a feasible sequence with the given budget"
            )
        if opt[cmem][lmin][lmax] == float("inf"):
            """
            print('a')
            print(chain.cw)
            print('abar')
            for i in range(chain.ln):
                print(chain.cbw[i])
                print(chain.fwd_tmp[i])
                print(chain.bwd_tmp[i])
            """
            raise ValueError(
                "Can't find a feasible sequence with the given budget"
                # f"Can't process this chain from index "
                # f"{lmin} to {lmax} with memory {memory_limit}"
            )

        if lmin == chain.ln:
            seq.insert(SeqLoss())
            return seq

        w = what[cmem][lmin][lmax]
        #  -- Solution 1 --
        if w[0]:
            k = w[1]
            sol = chain.body[lmin].sols[k]
            seq.insert(SeqBlockFe(lmin, sol.fwd_sched))
            seq.insert_seq(
                seq_builder_rec(lmin + 1, lmax, cmem - chain.cbw[lmin + 1][k])
            )
            seq.insert(SeqBlockBwd(lmin, sol.bwd_sched))

        #  -- Solution 1 --
        else:
            j = w[1]
            seq.insert(SeqBlockFc(lmin, chain.body[lmin].Fc_sched))
            for k in range(lmin + 1, j):
                seq.insert(SeqBlockFn(k, chain.body[k].Fn_sched))
            seq.insert_seq(seq_builder_rec(j, lmax, cmem - chain.cw[j]))
            seq.insert_seq(seq_builder_rec(lmin, j - 1, cmem))
        return seq

    #  ~~~~~~~~~~~~~~~~~~

    seq = seq_builder_rec(0, chain.ln, mmax)
    return seq


# ===================================
# =====  interface to C version =====
# ===================================

# The C version produces 'csequence' SeqOps, we have to convert them
import rockmate.csequence as cs

try:
    import rockmate.csolver as rs
    csolver_present = True
except:
    csolver_present = False


def convert_sequence_from_C(chain: RK_Chain, original_sequence):
    def convert_op(op):
        if isinstance(op, cs.SeqLoss):
            return SeqLoss()
        body = chain.body[op.index]
        if isinstance(op, cs.SeqBlockFn):
            return SeqBlockFn(op.index, body.Fn_sched)
        if isinstance(op, cs.SeqBlockFc):
            return SeqBlockFc(op.index, body.Fc_sched)
        if isinstance(op, cs.SeqBlockFe):
            return SeqBlockFe(op.index, body.sols[op.option].fwd_sched)
        if isinstance(op, cs.SeqBlockBwd):
            return SeqBlockBwd(op.index, body.sols[op.option].bwd_sched)

    result = RK_Sequence([convert_op(op) for op in original_sequence])
    return result


def csolve_dp_functionnal(chain: RK_Chain, mmax, opt_table=None):
    if opt_table is None:  ## TODO? if opt_table.mmax < mmax, create new table
        result = rs.RkTable(chain, mmax)
    else:
        result = opt_table
    result.get_opt(mmax)
    return result


def cseq_builder(chain: RK_Chain, mmax, opt_table):
    result = opt_table.build_sequence(mmax)
    return convert_sequence_from_C(chain, result)


# ===============================================
# =====  generic interface, selects version =====
# ===============================================


def solve_dp_functionnal(chain, mmax, opt_table=None, force_python=False):
    if force_python or not csolver_present:
        return psolve_dp_functionnal(chain, mmax, opt_table)
    else:
        return csolve_dp_functionnal(chain, int(mmax), opt_table)


def seq_builder(chain, mmax, opt_table):
    if csolver_present and isinstance(opt_table, rs.RkTable):
        return cseq_builder(chain, int(mmax), opt_table)
    else:
        return pseq_builder(chain, mmax, opt_table)




# ============
# = ROCKMATE =
# ============

import rkgb
from rkgb.main import make_inputs
from rkgb.utils import print_debug, np, irotor
from rkgb.utils.global_vars import ref_verbose
from rkgb.utils.small_fcts import get_device
from rkgb.utils.ast_add_on import ast_to_str
from rockmate.def_op import DelOp, OpSchedule
from rockmate.def_chain import RK_Chain
from rockmate.def_sequence import (
    SeqBlockBwd,
    SeqBlockFc,
    SeqBlockFn,
    SeqBlockFe,
)
from rockmate.rotor_solver import seq_builder, solve_dp_functionnal
from rockmate.translator import Translator, RngState
from rockmate.compiler import Compiler, RK_Storage
import torch
from torch import tensor
import ast
import time
import pickle
from datetime import datetime
import warnings
from os import environ


def print_memsizes(list_kg):
    di = list_kg[-1].dict_info
    for kg in list_kg:
        for n in kg.dict_nodes.values():
            mt = n.main_target
            try:
                print_debug(
                    f"{mt} : memsize {di[mt].memsize} ; " f"fm {n.fgt_mem}",
                    end="",
                )
            except:
                print_debug("\nloss")
    print_debug("\n")


class Rockmate(torch.nn.Module):
    def __init__(
        self,
        original_mod,
        model_inputs,
        budget=None,
        mem_unit=None,
        verbose=False,
        solve=True,
        get_sequence=True,
        get_compiled_fct=True,
        nb_budget_save=10,
        nb_budget_peak=5,
        ilp_solver="gurobi",
    ):
        super().__init__()
        ref_verbose[0] = verbose
        self.device = get_device()
        self.original_mod = original_mod
        self.mem_unit = mem_unit if mem_unit else 1024**2
        # -- use pytorch graph builder to get the list of K_graphs --
        self.rkgb_res = rkgb.make_all_graphs(
            original_mod, model_inputs, verbose=verbose, bool_kg=True
        )  # we don't need the whole K_graph
        self.list_kg = self.rkgb_res.K_graph_list
        self.dict_constants = self.rkgb_res.K_graph_list[0].dict_constants
        self.eq_classes = self.rkgb_res.equivalent_classes
        self.init_code = ast_to_str(self.list_kg[0].init_code)
        self.output = self.list_kg[-1].output_kdn_data
        self.budget = budget
        if solve:
            self.get_chain(nb_budget_save, nb_budget_peak)
            if get_sequence:
                self.get_sequence(budget)
                if get_compiled_fct:
                    self.get_compiled_fct()

    def get_chain(self, nb_budget_save=10, nb_budget_peak=5):
        start = time.time()
        #  -- use checkmate to solve all the blocks --
        self.rk_chain = RK_Chain(
            self.list_kg,
            self.eq_classes,
            nb_budget_save,
            nb_budget_peak,
            mem_unit=self.mem_unit,
        )
        end = time.time()
        self.ILP_solve_time = end - start

        self.opt_table = None

    def get_sequence(self, budget):
        for n, p in self.original_mod.named_parameters():
            if p.grad is None:
                p.grad = torch.zeros_like(p)

        self.peak_overhead = max(
            [
                sum(
                    [
                        kdn.mem
                        for kdn in kcn.deps_fake
                        if kdn.main_target != kcn.main_target
                    ]
                )
                for kg in self.list_kg
                for kcn in kg.list_kcn
            ]
        )

        if budget:
            self.budget = budget
        else:
            self.budget = (
                # If not given a budget, we use the current available memory
                torch.cuda.get_device_properties(0).total_memory * 0.9
                - torch.cuda.memory_allocated()
                - self.peak_overhead
            )
        print_debug("budget", self.budget)
        # -- solve the chain like rotor --
        start = time.time()
        mmax = self.budget // self.mem_unit - self.rk_chain.cw[0]
        self.opt_table = solve_dp_functionnal(self.rk_chain, mmax, self.opt_table)
        self.seq = seq_builder(
            self.rk_chain, self.budget // self.mem_unit, self.opt_table
        )
        end = time.time()
        self.DP_solve_time = end - start

        enable = np.zeros(len(self.list_kg))
        for s in self.seq.seq:
            if isinstance(s, SeqBlockFe):
                enable[s.index] = 1
            if isinstance(s, SeqBlockBwd):
                if not enable[s.index - 1]:
                    s.op_sched.del_input()
        self.fwd_seq, self.bwd_seq = self.seq.cut_fwd_bwd()
        self.fwd_op_sched_list = [seq.op_sched for seq in self.fwd_seq.seq]
        self.bwd_op_sched_list = [seq.op_sched for seq in self.bwd_seq.seq]

        self.fwd_op_list = [
            op for op_sched in self.fwd_op_sched_list for op in op_sched.op_list
        ]
        self.bwd_op_list = [
            op for op_sched in self.bwd_op_sched_list for op in op_sched.op_list
        ]

        self.fwd_alive_list = []
        self.bwd_alive_list = []
        list_kdn = []
        start_i = []
        for kg in self.list_kg:
            start_i.append(len(list_kdn))
            list_kdn += kg.list_kdn
        alive_status = np.zeros(len(list_kdn) + 2, dtype=bool)
        alive_status[-1] = 1
        kdn_names = [kdn.name for kdn in list_kdn] + [
            self.list_kg[0].input_kdn_grad.name,
            self.list_kg[0].input_kdn_data.name,
        ]
        for op_sched in self.fwd_op_sched_list:
            index = [kdn_names.index(n) for n in op_sched.kdn_names]
            for a in op_sched.alive_list:
                alive_status[index] = a
                self.fwd_alive_list.append(alive_status.copy())
        for op_sched in self.bwd_op_sched_list:
            index = [kdn_names.index(n) for n in op_sched.kdn_names]
            for a in op_sched.alive_list:
                alive_status[index] = a
                self.bwd_alive_list.append(alive_status.copy())

        self.fwd_op_sched = OpSchedule(
            self.fwd_op_list,
            self.fwd_alive_list,
            self.list_kg[0].input_kdn_data,
            self.list_kg[0].input_kdn_grad,
            self.output,
            list_kdn,
        )

        self.bwd_op_sched = OpSchedule(
            self.bwd_op_list,
            self.bwd_alive_list,
            self.list_kg[0].input_kdn_data,
            self.list_kg[0].input_kdn_grad,
            self.output,
            list_kdn,
        )

        self.op_sched = OpSchedule(
            self.fwd_op_list + self.bwd_op_list,
            self.fwd_alive_list + self.bwd_alive_list,
            self.list_kg[0].input_kdn_data,
            self.list_kg[0].input_kdn_grad,
            self.output,
            list_kdn,
        )

        self.simulation_time = sum(op.time for op in self.fwd_op_list) + sum(
            op.time for op in self.bwd_op_list
        )

        self.simulation_overhead = self.simulation_time / sum(
            [kcn.time for kg in self.list_kg for kcn in kg.list_kcn]
        )
        self.storage = RK_Storage(self.device, self.original_mod, self.dict_constants)

    def get_compiled_fct(self):
        self.compiler = Compiler(self.storage)
        self.fct_list = self.compiler.compile(self.op_sched)
        loss_idx = len(self.fwd_op_list)
        self.fwd_fct_list = self.fct_list[:loss_idx]
        self.bwd_fct_list = self.fct_list[loss_idx:]

    def _exec(self, fct_list, record_mem=False, compiled=False):
        if not compiled:
            warnings.warn("Translator is no longer used!")
        if record_mem:
            torch.cuda.reset_peak_memory_stats()
            self.mem_before = torch.cuda.memory_allocated()
            self.max_before = torch.cuda.max_memory_allocated()
            for fct in fct_list:
                fct()
            allo_mem = torch.cuda.memory_allocated() - self.mem_before
            peak_mem = torch.cuda.max_memory_allocated() - self.max_before
            self.max_mem.append(peak_mem - allo_mem)
            self.allo_mem.append(allo_mem)
        else:
            for fct in fct_list:
                fct()

    def forward(self, *args, record_mem=False, compiled=True, **kwargs):
        if not self.training:
            self.original_mod.eval()
            return self.original_mod(*args, **kwargs)
        model_inputs = make_inputs(self.original_mod, args, kwargs)
        for k, v in model_inputs.items():
            self.storage.add_val(k, v)
        exec(self.init_code, self.storage.gd, self.storage.ld)
        for kg in self.list_kg:
            for kdn in kg.list_kdn:
                self.storage.ld[kdn.main_target] = torch.empty(
                    0, device=self.device, requires_grad=kdn.info.requires_grad
                )
        self.max_mem = []
        self.allo_mem = []
        if compiled:
            for l in self.fwd_fct_list:
                self._exec(l, record_mem, compiled=compiled)
            return self.storage.get_val(self.output.main_target)

        return self.storage.get_val(self.output.main_target)

    def backward(
        self, stop=False, record_mem=False, add_output_grad=True, compiled=True
    ):
        if record_mem:
            self.output_size = irotor.tensorMsize(
                self.storage.ld[self.output.main_target]
            )
            # self.allo_mem[-1] += self.output.info.memsize
            # output grad is generated outside
            loss_idx = len(self.allo_mem)
        if stop:
            for l in self.bwd_fct_list[: stop - len(self.fwd_fct_list)]:
                self._exec(l, record_mem, compiled=compiled)
            return None
        if compiled:
            for l in self.bwd_fct_list:
                self._exec(l, record_mem, compiled=compiled)

        if record_mem and add_output_grad:
            self.allo_mem[loss_idx] += self.output_size

    def expect_time(self):
        # Sum of the measured time of each operation for one batch
        return self.fwd_seq.compute_time() + self.bwd_seq.compute_time()

    def expect_mem(self, overhead=False):
        # Peak mem based on the measured memory/overhead of each operation
        pred_mem = []
        acc_mem = np.zeros(len(self.fwd_seq.seq))
        # alive_dict = {}
        # for kg in self.list_kg:
        #     for kdn in kg.list_kdn:
        #         alive_dict[kdn.name] = (0, kdn.mem)
        for i, seq in enumerate(self.fwd_seq.seq + self.bwd_seq.seq):
            op_sched = seq.op_sched
            for a, op in zip(op_sched.alive_list, op_sched.op_list):
                acc_mem[seq.index] = (
                    np.dot(a, op_sched.mem_sizes) - op_sched.input_size[1]
                )
                # if not op_sched.is_fwd:
                #     acc_mem[seq.index] -= op_sched.output_size[1]
                pred_mem.append(sum(acc_mem))
                if overhead and op.op_type == "Run":
                    pred_mem[-1] += op.overhead
            # for s, op in zip(op_sched.save, op_sched.op_list):
            # for i, op in enumerate(op_sched.op_list):
            # acc_mem[seq.index] = s
            # pred_mem.append(sum(acc_mem))
            # if overhead and op.op_type == "Run":
            #     pred_mem[-1] += op.overhead
        return pred_mem

    def reinit(self):
        self.original_mod.zero_grad()
        self.storage.ld = {}

    def save_to_file(self, path, id=datetime.now().strftime("%d_%m_%Y_%H_%M_%S")):
        sol = {}
        sol["op_sched"] = self.op_sched
        sol["loss_idx"] = len(self.fwd_op_list)
        with open(f"{path}/{id}_solution.pkl", "wb") as f:
            pickle.dump(sol, f)

    def load_from_file(self, path, id=datetime.now().strftime("%d_%m_%Y_%H_%M_%S")):
        with open(f"{path}/{id}_solution.pkl", "rb") as f:
            sol = pickle.load(f)
        op_sched = sol["op_sched"]
        loss_idx = sol["loss_idx"]
        self.storage = RK_Storage(self.device, self.original_mod, self.dict_constants)
        self.compiler = Compiler(self.storage)
        self.fct_list = self.compiler.compile(op_sched)
        self.fwd_fct_list = self.fct_list[:loss_idx]
        self.bwd_fct_list = self.fct_list[loss_idx:]



from rkgb.utils.ast_add_on import make_str_assign, make_str_list_assign
from rkgb.utils import np, torch
from rockmate.def_op import DelOp


class RngState:
    def __init__(self):
        self.cpu_states = {}
        self.gpu_states = {}

    def get(self, op_name):
        if op_name not in self.cpu_states.keys():
            self.cpu_states[op_name] = torch.get_rng_state()
            self.gpu_states[op_name] = torch.cuda.get_rng_state()

    def restore(self, op_name):
        # pass
        torch.set_rng_state(self.cpu_states[op_name])
        torch.cuda.set_rng_state(self.gpu_states[op_name])


class Translator:  # to execute Op
    def __init__(self, storage, aggressive=True):
        self.storage = storage
        self.live = {}
        self.fgt = []
        self.code = []
        self.grad = {}
        self.fwd_op_sched = []
        self.bwd_op_sched = []
        self.op_info = []
        self.fwd_code = []
        self.aggressive = aggressive
        if self.aggressive:
            self.alive_global = {}
            self.info_global = {}

    def _estimate_memory(self):
        mem = 0
        for k, v in self.live.items():
            mt, data = k.split(".")
            if v:
                mem += self.mt2op[mt].mem
        return mem

    def translate(self, op_sched, during_fwd=True, first=False):
        if self.aggressive:
            for i, kdn_name in enumerate(op_sched.kdn_names):
                self.alive_global[kdn_name] = op_sched.alive_list[-1][i]
                self.info_global[kdn_name] = op_sched.kdn_info[kdn_name]
        else:
            self.alive_global = {}
        op_name_list = [op.name for op in op_sched.op_list]
        # Fc/Fn cases
        if op_sched.no_grad:
            code_list = []  # ["with torch.no_grad():"]
            for i, op in enumerate(op_sched.op_list):
                if op.op_type == "Run":
                    if "loss" in op.main_target:
                        code_list.append("")
                    else:
                        # code = ast_to_str(make_ast_module([op.main_code]))
                        # code += "\n"+ast_to_str(make_ast_module(op.body_code))
                        # code = op.code
                        # code = "\t".join(code.splitlines(True))
                        if op.is_rand:
                            code = f"rng_state.get('{op.name}');rng_state.restore('{op.name}')\n{op.ff_code}"
                        else:
                            code = op.ff_code
                        # for target in op.tensor_targets:
                        #     code += f"shapes['{target}'] = {target}.shape;"
                        # for phantom_name in op.phantom_names:
                        #     code += (
                        #         f"shapes['{phantom_name}'] = _{phantom_name}.shape;"
                        #     )
                        code_list.append(f"{code}")

                elif op.kdn_type == "data":
                    code = ""
                    if op_sched.del_input_idx == i:
                        for target in op_sched.del_input_op.tensor_targets:
                            # code += f"del {target};"
                            code += (
                                f"{target}.data = torch.empty(0,device=device);"
                            )
                    else:
                        for target in op.tensor_targets:
                            code += (
                                f"{target}.data = torch.empty(0,device=device);"
                            )
                    if op.includes_base:
                        code += f"{op.main_target}._base.data = torch.empty(0,device=device);"

                        # code += f"del {target};"
                    code_list.append(code)
                else:
                    code_list.append("")
                # if op_sched.del_input_idx == i:
                #     code = "\n"
                #     for target in op_sched.del_input_op.tensor_targets:
                #         code += f"{target}.data = torch.empty(0,device=device);"
                #     code_list[-1] += code
            out_target = op_sched.output_size[0]
            code_list[-1] += f"\n{out_target}.requires_grad_();"
            code_list[-1] += f"shapes['{out_target}'] = {out_target}.shape;"
            # if first:
            #     out_op = [
            #         op for op in op_sched.op_list if out_target in op.name
            #     ][0]
            #     for target in out_op.tensor_targets:
            #         if "loss" not in target:
            #             code_list[-1] += f"shapes['{target}'] = {target}.shape;"
            #     for phantom_name in out_op.phantom_names:
            #         code_list[
            #             -1
            #         ] += f"shapes['{phantom_name}'] = _{phantom_name}.shape;"
            return code_list

        def _is_alive(kdn_name, i):
            if kdn_name in op_sched.kdn_names:
                return op_sched.alive_list[i][
                    op_sched.kdn_names.index(kdn_name)
                ]
            elif kdn_name in self.alive_global:
                return self.alive_global[kdn_name]
            else:
                return True

        def _is_proxy_alive(kdn_target, i):
            if f"{kdn_target} data" not in op_sched.kdn_dict:
                # belong to the next block, or del_input
                return False
            includes_phantoms = op_sched.kdn_dict[
                f"{kdn_target} data"
            ].includes_phantoms

            if f"bwd_{kdn_target}" in op_name_list[:i]:
                return False
            if f"bwd_{kdn_target}" in op_name_list[i]:
                return True
            if f"{kdn_target} phantoms" in op_sched.kdn_names:
                return op_sched.alive_list[i][
                    op_sched.kdn_names.index(f"{kdn_target} phantoms")
                ]
            elif (
                includes_phantoms
            ):  # phantom kdn does not exist, need to check data kdn
                #  op_sched.op_list
                return op_sched.alive_list[i][
                    op_sched.kdn_names.index(f"{kdn_target} data")
                ]
            else:
                return f"fwd_{kdn_target}" in op_name_list[:i] or (
                    not op_sched.is_fwd
                )
            #     if op_sched.is_fwd:
            #         if kdn_target == op_sched.input_size[0]:
            #             # input was generated previously
            #             return True
            #         return f"fwd_{kdn_target}" in op_name_list[:i]
            #     else:
            #         return f"bwd_{kdn_target}" in op_name_list[i:]

        def _generate_fake_data(kdn, i, is_self=False):
            # return code for generate the target fake tensor (only for data/grad)
            prep_code = ""
            after_code = ""
            req_shape = kdn.info.tsize
            target_tensor = None
            mt = kdn.main_target
            dict_info = (
                self.info_global if self.aggressive else op_sched.kdn_info
            )

            target_tensor = f"metensor.clone().expand(np.prod(shapes['{mt}']))"
            prep_code += f"{mt}.data = {target_tensor}.view(shapes['{mt}']);"

            for v in kdn.tensor_targets:
                after_code += f"{v}.data = torch.empty(0,device=device); "
            if is_self:
                prep_code += (
                    f"_{mt}.data = {target_tensor}.view(shapes['{mt}']);"
                )
                after_code += f"_{mt}.data = torch.empty(0,device=device);"
            return prep_code, after_code

        def _run_op(op, i):
            # Forward operation
            mt = op.main_target
            if "fwd" in op.name:
                rec = (i > op_sched.op_list.index(op)) or (not op_sched.is_fwd)
                force = rec or not first
                suffix = ""
                if rec and not op.proxy and "loss" not in op.name:
                    suffix = ".data"
                code = (
                    make_str_assign(
                        op.main_code, suffix=suffix, force_special_kwargs=force
                    )
                    + "\n"
                )

                if op.proxy:
                    if (
                        (not during_fwd)
                        and (not op_sched.no_grad)
                        and (mt == op_sched.output_size[0])
                    ):
                        rec = True
                    # code = make_str_assign(op.main_code, prefix="_") + ";"
                    # if not rec:
                    #     code += f"{mt} = _{mt};\n"

                # else:
                #     code = make_str_assign(op.main_code) + "\n"
                code += (
                    make_str_list_assign(
                        op.inplace_code, force_special_kwargs=force
                    )
                    + "\n"
                )
                if op.proxy:
                    for target in op.tensor_targets:
                        code = code.replace(target, "_" + target)
                    if rec:
                        code += f"{mt}.data = _{mt}.data;\n"
                    else:
                        code += (
                            f"{mt} = _{mt}.detach();{mt}.requires_grad_();\n"
                        )
                for bc in op.body_code:
                    suffix = ""
                    if rec and (bc[0] in op.tensor_targets):
                        suffix = ".data"
                    code += (
                        make_str_assign(
                            bc, suffix=suffix, force_special_kwargs=force
                        )
                        + "\n"
                    )
                rec = (i > op_sched.op_list.index(op)) or (not op_sched.is_fwd)
                for user, tensor, phantom_name in op.alias_in_users_phantoms:
                    if rec and _is_proxy_alive(user, i):
                        code += f"_{phantom_name}.data = {tensor}.view(shapes['{phantom_name}']);"

                if not rec:
                    for target in op.tensor_targets:
                        if "loss" not in target:
                            code += f"shapes['{target}'] = {target}.shape;"
                    for phantom_name in op.phantom_names:
                        code += (
                            f"shapes['{phantom_name}'] = _{phantom_name}.shape;"
                        )
                if op.is_rand:
                    code = f"rng_state.get('{op.name}');rng_state.restore('{op.name}')\n{code}"
                return code
            # Backward operation
            elif "bwd" in op.name:
                mt = op.main_target
                rec = op in op_sched.op_list[:i]
                last = not (op in op_sched.op_list[i + 1 :])
                prep_code = ""
                after_code = ""
                for kdn in op.deps_fake:
                    if (
                        not _is_alive(kdn.name, i)
                        # or op_sched.input_size[0] in kdn.name
                    ):
                        fake_code = _generate_fake_data(
                            kdn, i, is_self=(kdn.main_target == op.main_target)
                        )
                        prep_code += fake_code[0]
                        after_code += fake_code[1]
                if rec:
                    prev_i = i - op_sched.op_list[:i][::-1].index(op) - 1
                    rec_list = []
                    for kdn in op.users_global:
                        if DelOp(kdn) in op_sched.op_list[prev_i:i]:
                            rec_list += [kdn.main_target]  # kdn.tensor_targets
                    inputs = ",".join(rec_list)
                    code = f"_{mt}.backward({mt}.grad, inputs=[{inputs}], retain_graph={not last})"
                else:
                    code = f"_{mt}.backward({mt}.grad, retain_graph={not last})"
                bwd_code = f"{prep_code}\n" f"{code}\n" f"{after_code}"
                if op.is_rand:
                    bwd_code = f"rng_state.get('{op.name}');rng_state.restore('{op.name}')\n{bwd_code}"
                return bwd_code

        def _del_op(op, i):
            code = ""
            if op.kdn_type == "data":
                for user, tensor, phantom_name in op.alias_in_users_phantoms:
                    if _is_proxy_alive(user, i):
                        code += f"_{phantom_name}.data = torch.empty(0,device=device); "
                if (
                    op.info is not None
                    and op.info.requires_grad
                    and _is_alive(op.name.replace("data", "phantoms"), i)
                    and op.proxy
                ):
                    code += f"_{op.main_target}.data = torch.empty(0,device=device);"
                    for inp in op.inplace_targets:
                        # code += f"_{inp}.data = torch.empty(0,device=device);"
                        code += f"del _{inp};"

                    if op.includes_phantoms:
                        code += f"del _{op.main_target};"
                    if op.includes_base:
                        if op.proxy:
                            code += f"_{op.main_target}._base.data = torch.empty(0,device=device);"
                        else:
                            code += f"{op.main_target}._base.data = torch.empty(0,device=device);"

                for v in op.tensor_targets:
                    code += f"{v}.data = torch.empty(0,device=device); "

                for v in op.container_targets:
                    code += f"del {v};"

            if op.kdn_type == "grad":
                code += f"{op.main_target}.grad = None;"
            if op.kdn_type == "phantoms":
                code += f"del _{op.main_target};"
                # for inp in op_sched.kdn_dict[f"{op.main_target} phantoms"].inplace_targets:
                #     code += f"del _{inp};"
            return code

        code_list = []
        for i, (op, alive) in enumerate(
            zip(op_sched.op_list, op_sched.alive_list)
        ):
            if op.op_type == "Run":
                code_list.append(_run_op(op, i))
            if op.op_type == "Del":
                code_list.append(_del_op(op, i))
            # if op_sched.del_input_idx == i:
            #     code = "\n"
            #     for target in op_sched.del_input_op.tensor_targets:
            #         code += f"{target}.data = torch.empty(0,device=device);"
            #     code_list[-1] += code
        return code_list



import torch
import math
import copy
import torch.nn as nn
import torch.nn.functional as F
from torch.nn.modules import ModuleList
from torch.nn.modules.normalization import LayerNorm


class Conv1D(nn.Module):
    def __init__(self, nx, nf):
        super().__init__()
        self.nf = nf
        w = torch.empty(nx, nf)
        nn.init.normal_(w, std=0.02)
        self.weight = nn.Parameter(w)
        self.bias = nn.Parameter(torch.zeros(nf))

    def forward(self, x):
        size_out = x.size()[:-1] + (self.nf,)
        x = torch.addmm(self.bias, x.view(-1, x.size(-1)), self.weight)
        x = x.view(size_out)
        return x


class FeedForward(nn.Module):
    def __init__(self, dropout, d_model=768, nx=768 * 4):
        super().__init__()
        self.c_fc = Conv1D(d_model, nx)
        self.c_proj = Conv1D(nx, d_model)
        self.act = F.gelu
        self.dropout = nn.Dropout(dropout)
        # self.dropout = nn.Identity()

    def forward(self, x):
        return self.dropout(self.c_proj(self.act(self.c_fc(x))))


class Attention(nn.Module):
    def __init__(
        self,
        d_model=768,
        n_head=12,
        n_ctx=1024,
        d_head=64,
        bias=True,
        scale=False,
        dropout=0.1,
    ):
        super().__init__()
        self.n_head = n_head
        self.d_model = d_model
        self.c_attn = Conv1D(d_model, d_model * 3)
        self.scale = scale
        self.softmax = nn.Softmax(dim=-1)
        self.register_buffer(
            "bias",
            torch.tril(torch.ones(n_ctx, n_ctx)).view(1, 1, n_ctx, n_ctx),
        )
        self.dropout = nn.Dropout(dropout)
        # self.dropout = nn.Identity()
        self.c_proj = Conv1D(d_model, d_model)

    def split_heads(self, x):
        "return shape [`batch`, `head`, `sequence`, `features`]"
        new_shape = x.size()[:-1] + (self.n_head, x.size(-1) // self.n_head)
        x = x.view(new_shape)
        return x.permute(0, 2, 1, 3)

    def _attn(self, q, k, v, attn_mask=None):
        scores = torch.matmul(q, k.transpose(-2, -1))
        if self.scale:
            scores = scores / math.sqrt(v.size(-1))
        nd, ns = scores.size(-2), scores.size(-1)
        if attn_mask is not None:
            scores = scores + attn_mask
        scores = self.softmax(scores)
        scores = self.dropout(scores)
        outputs = torch.matmul(scores, v)
        return outputs

    def merge_heads(self, x):
        x = x.permute(0, 2, 1, 3).contiguous()
        new_shape = x.size()[:-2] + (x.size(-2) * x.size(-1),)
        return x.view(new_shape)

    def forward(self, x):
        x = self.c_attn(x)  # new `x` shape - `[1,3,2304]`
        q, k, v = x.split(self.d_model, dim=2)
        q, k, v = self.split_heads(q), self.split_heads(k), self.split_heads(v)
        # out      = self._attn(q, k, v)
        scores = torch.matmul(q, k.transpose(-2, -1))
        # if self.scale: scores = scores/math.sqrt(v.size(-1))
        nd, ns = scores.size(-2), scores.size(-1)
        # if attn_mask is not None: scores = scores + attn_mask
        scores = self.softmax(scores)
        scores = self.dropout(scores)
        out = torch.matmul(scores, v)

        out = self.merge_heads(out)
        out = self.c_proj(out)
        return out


# class TransformerBlock(nn.Module):
#     def __init__(self, d_model=768, n_head=12, dropout=0.1):
#         super(TransformerBlock, self).__init__()
#         self.attn = Attention(
#             d_model=d_model,
#             n_head=n_head,
#             d_head=64,
#             n_ctx=1024,
#             bias=True,
#             scale=False,
#             dropout=dropout,
#         )
#         self.feedforward = FeedForward(
#             dropout=dropout, d_model=d_model, nx=d_model * 4
#         )
#         self.ln_1 = LayerNorm(d_model)
#         self.ln_2 = LayerNorm(d_model)

#     def forward(self, x):
#         x1 = self.ln_1(x)
#         x2 = self.ln_2(x)
#         x = x + self.attn(x1)
#         x = x + self.feedforward(x2)
#         return x


class Residual(nn.Module):
    def __init__(self, layer1, layer2):
        super(Residual, self).__init__()
        self.layer1 = layer1
        self.layer2 = layer2

    def forward(self, x):
        return x + self.layer2(self.layer1(x))


def getTransformer(d_model=768, n_head=12, dropout=0.1):
    attn = Attention(
        d_model=d_model,
        n_head=n_head,
        d_head=64,
        n_ctx=1024,
        bias=True,
        scale=False,
        dropout=dropout,
    )
    feedforward = FeedForward(dropout=dropout, d_model=d_model, nx=d_model * 4)
    ln1 = LayerNorm(d_model)
    ln2 = LayerNorm(d_model)
    return nn.Sequential(*[Residual(ln1, attn), Residual(ln2, feedforward)])


class TransformerBlock(nn.Module):
    def __init__(self, d_model=768, n_head=12, dropout=0.1):
        super(TransformerBlock, self).__init__()
        self.attn = Attention(
            d_model=d_model,
            n_head=n_head,
            d_head=64,
            n_ctx=1024,
            bias=True,
            scale=False,
            dropout=dropout,
        )
        self.feedforward = FeedForward(
            dropout=dropout, d_model=d_model, nx=d_model * 4
        )
        self.ln_1 = LayerNorm(d_model)
        self.ln_2 = LayerNorm(d_model)

    def forward(self, x):
        x1 = self.ln_1(x)
        x2 = self.ln_2(x)
        x = x + self.attn(x1)
        x = x + self.feedforward(x2)
        return x


def _get_clones(module, n):
    return ModuleList([copy.deepcopy(module) for i in range(n)])


class GPT2(nn.Module):
    def __init__(
        self,
        nlayers=12,
        n_ctx=2048,
        n_head=12,
        d_model=768,
        vcb_sz=50257,
        dropout=0.1,
    ):
        super(GPT2, self).__init__()
        self.nlayers = nlayers
        # block = TransformerBlock(
        #     d_model=d_model, n_head=n_head, dropout=dropout
        # )
        block = getTransformer(d_model=d_model, n_head=n_head, dropout=dropout)
        self.h = _get_clones(block, nlayers)
        self.wte = nn.Embedding(vcb_sz, d_model)
        self.wpe = nn.Embedding(n_ctx, d_model)
        self.drop = nn.Dropout(dropout)
        # self.drop = nn.Identity()

        self.ln_f = LayerNorm(d_model)
        self.out = nn.Linear(d_model, vcb_sz, bias=False)
        self.loss_fn = nn.CrossEntropyLoss()
        self.init_weights()
        self.n_head = n_head
        self.d_model = d_model

    def init_weights(self):
        self.out.weight = self.wte.weight
        self.apply(self._init_weights)

    def _init_weights(self, module):
        if isinstance(module, (nn.Linear, nn.Embedding, Conv1D)):
            module.weight.data.normal_(mean=0.0, std=0.02)
            if (
                isinstance(module, (nn.Linear, Conv1D))
                and module.bias is not None
            ):
                module.bias.data.zero_()
        elif isinstance(module, nn.LayerNorm):
            module.bias.data.zero_()
            module.weight.data.fill_(1.0)

    def forward(
        self, src, labels=None, pos_ids=None, return_inp=False, dropout=0.1
    ):
        if pos_ids is None:
            pos_ids = torch.arange(
                0, src.size(-1), dtype=torch.long, device=self.wpe.weight.device
            ).unsqueeze(0)
        inp = self.drop((self.wte(src) + self.wpe(pos_ids)))
        if return_inp:
            return inp
        for i in range(self.nlayers):
            inp = self.h[i](inp)
        inp = self.ln_f(inp)
        logits = self.out(inp)
        outputs = (logits,) + (inp,)

        if labels is not None:
            shift_logits = logits[..., :-1, :].contiguous()
            shift_labels = labels[..., 1:].contiguous()
            loss = self.loss_fn(
                shift_logits.view(-1, shift_logits.size(-1)),
                shift_labels.view(-1),
            )
            outputs = (loss,) + outputs
            return outputs
        return logits


class GPT2_input(nn.Module):
    def __init__(self, n_ctx=1024, d_model=768, vcb_sz=50257, dropout=0.1):
        super(GPT2_input, self).__init__()
        block = TransformerBlock(d_model=d_model, n_head=12, dropout=dropout)
        # self.h = _get_clones(block, nlayers)
        self.wte = nn.Embedding(vcb_sz, d_model)
        self.wpe = nn.Embedding(n_ctx, d_model)
        self.drop = nn.Dropout(dropout)
        # self.drop = nn.Identity()
        self.ln_f = LayerNorm(d_model)
        self.out = nn.Linear(d_model, vcb_sz, bias=False)
        self.loss_fn = nn.CrossEntropyLoss()
        self.init_weights()

    def init_weights(self):
        self.out.weight = self.wte.weight
        self.apply(self._init_weights)

    def _init_weights(self, module):
        if isinstance(module, (nn.Linear, nn.Embedding, Conv1D)):
            module.weight.data.normal_(mean=0.0, std=0.02)
            if (
                isinstance(module, (nn.Linear, Conv1D))
                and module.bias is not None
            ):
                module.bias.data.zero_()
        elif isinstance(module, nn.LayerNorm):
            module.bias.data.zero_()
            module.weight.data.fill_(1.0)

    def forward(
        self, src, labels=None, pos_ids=None, return_inp=False, dropout=0.1
    ):
        if pos_ids is None:
            pos_ids = torch.arange(
                0, src.size(-1), dtype=torch.long, device=self.wpe.weight.device
            ).unsqueeze(0)
        inp = self.drop((self.wte(src) + self.wpe(pos_ids)))
        return inp


class GPT2_output(nn.Module):
    def __init__(self, d_model=768, vcb_sz=50257, dropout=0.1):
        super(GPT2_output, self).__init__()
        # self.nlayers = nlayers
        # block = TransformerBlock(d_model=d_model, n_head=12, dropout=dropout)
        # self.h = _get_clones(block, nlayers)
        self.wte = nn.Embedding(vcb_sz, d_model)
        # self.wpe = nn.Embedding(n_ctx, d_model)
        self.drop = nn.Dropout(dropout)
        # self.drop    = nn.Identity()
        self.ln_f = LayerNorm(d_model)
        self.out = nn.Linear(d_model, vcb_sz, bias=False)
        self.loss_fn = nn.CrossEntropyLoss()
        self.init_weights()

    def init_weights(self):
        self.out.weight = self.wte.weight
        self.apply(self._init_weights)

    def _init_weights(self, module):
        if isinstance(module, (nn.Linear, nn.Embedding, Conv1D)):
            module.weight.data.normal_(mean=0.0, std=0.02)
            if (
                isinstance(module, (nn.Linear, Conv1D))
                and module.bias is not None
            ):
                module.bias.data.zero_()
        elif isinstance(module, nn.LayerNorm):
            module.bias.data.zero_()
            module.weight.data.fill_(1.0)

    def forward(
        self, inp, labels=None, pos_ids=None, return_inp=False, dropout=0.1
    ):
        inp = self.ln_f(inp)
        logits = self.out(inp)
        outputs = (logits,) + (inp,)

        if labels is not None:
            shift_logits = logits[..., :-1, :].contiguous()
            shift_labels = labels[..., 1:].contiguous()
            loss = self.loss_fn(
                shift_logits.view(-1, shift_logits.size(-1)),
                shift_labels.view(-1),
            )
            outputs = (loss,) + outputs
            return outputs
        return logits


def get_GPT(model="GPT2-small"):
    if model == "GPT2-small":  # 117M parameters
        return GPT2(nlayers=12, d_model=768, n_head=12)
    if model == "GPT2-medium":  # 345M parameters
        return GPT2(nlayers=24, d_model=1024, n_head=16)
    if model == "GPT2-large":  # 774M parameters
        return GPT2(nlayers=36, d_model=1280, n_head=20)
    if model == "GPT2-xl":  # 1.6B parameters
        return GPT2(nlayers=48, d_model=1600, n_head=25)

    if model == "GPT3-small":  # 125M parameters
        return GPT2(nlayers=12, d_model=768, n_head=12)
    if model == "GPT3-medium":  # 350M parameters
        return GPT2(nlayers=24, d_model=1024, n_head=16)
    if model == "GPT3-large":  # 760M parameters
        return GPT2(nlayers=24, d_model=1536, n_head=16)
    if model == "GPT3-xl":  # 1.3B parameters
        return GPT2(nlayers=24, d_model=2048, n_head=24)




# ==========================
# everything about codes, how
# we exec it and how we store things
# -> to replace rotor/Checkpointable.functions
# ==========================
from rkgb.utils.small_fcts import check_attr
import numpy as np
import torch
import warnings


class RunOp:
    def __init__(self, kcn, keep_kcn=False):
        self.name = kcn.name
        self.time = kcn.time
        self.overhead = kcn.overhead
        self.main_target = kcn.main_target
        self.tensor_targets = kcn.tensor_targets
        # self.save_mem = cn.mem
        self.main_code = kcn.main_code
        self.body_code = kcn.body_code
        self.inplace_code = kcn.inplace_code
        self.ff_code = kcn.get_code(force_special_kwargs=True)
        self.deps_fake = [kdn.name for kdn in kcn.deps_fake]
        self.deps_global = [kdn.name for kdn in kcn.deps_global]
        self.users_global = [kdn.name for kdn in kcn.users_global]
        # self.deps_global = kcn.deps_global
        # self.deps_fake = kcn.deps_fake
        # self.users_global = kcn.users_global
        # self.alias_in_users_phantoms = kcn.alias_in_users_phantoms
        # self.phantom_names = kcn.phantom_names
        if keep_kcn:
            self.kcn = kcn
        self.is_fgt = False
        self.op_type = "Run"
        self.proxy = False
        self.no_grad = False
        for kdn in kcn.users:
            if kdn.kdn_type != "data":
                continue
            self.proxy = kdn.info.requires_grad
        self.is_rand = kcn.is_rand

    def __eq__(self, op2):
        return check_attr(self, op2, ["name"])

    def __str__(self):
        return f"Run {self.name}"


class DelOp:
    def __init__(self, kdn, proxy=True):
        self.name = kdn.name
        self.kdn_type = kdn.kdn_type
        self.time = 0
        self.save_mem = kdn.mem
        self.main_target = kdn.main_target
        # self.tensor_targets = kdn.all_targets
        # self.all_targets = list(kdn.deps)[0].all_targets if kdn.deps else []
        self.tensor_targets = kdn.tensor_targets
        self.all_targets = kdn.all_targets
        self.container_targets = kdn.container_targets
        self.inplace_targets = kdn.inplace_targets
        # self.code = kn.get_code()
        # self.requires_grad = kdn.info.requires_grad
        self.info = kdn.info
        self.is_fgt = True
        self.op_type = "Del"
        self.proxy = proxy
        self.includes_phantoms = kdn.includes_phantoms
        self.includes_base = kdn.includes_base
        # self.alias_in_users_phantoms = kdn.alias_in_users_phantoms

    def __eq__(self, op2):
        return check_attr(self, op2, ["name"])

    def __str__(self):
        return f"Del {self.name}"


class OpSchedule:
    def __init__(
        self,
        op_list,
        alive_list,
        input_kdn_data,
        input_kdn_grad,
        output_kdn_data,
        list_kdn,
        no_grad=False,
    ):
        self.op_list = op_list
        self.op_name_list = [op.name for op in self.op_list]
        self.alive_list = alive_list
        L = len(op_list)

        self.no_grad = no_grad

        self.input_size = (
            input_kdn_data.main_target,
            input_kdn_data.mem,
        )
        self.output_size = (
            output_kdn_data.main_target,
            output_kdn_data.mem,
        )
        self.kdn_dict = {kdn.name: kdn for kdn in list_kdn}

        # save the del_input op in case needed
        input_kdn = input_kdn_data
        self.del_input_op = DelOp(input_kdn, proxy=False)
        self.del_input_idx = L

        list_kdn = list_kdn + [input_kdn_grad, input_kdn_data]
        self.mem_sizes = [kdn.mem for kdn in list_kdn]
        self.kdn_names = [kdn.name for kdn in list_kdn]
        self.kdn_info = {
            kdn.name: kdn.info for kdn in list_kdn
        }  # dict: name->info

        self.is_fwd = True
        self.get_mem_time()
        assert self.valid_sched()

    def get_mem_time(self):
        """
        everytime op_list/alive_list are changed, run this to update mem
        """
        L = len(self.op_list)
        self.save = np.zeros(L)
        self.tmp = np.zeros(L)
        input_grad = False
        output_grad = False
        for i, op in enumerate(self.op_list):
            if isinstance(op, RunOp):
                self.tmp[i] = op.overhead
                if "bwd" in op.name:
                    self.is_fwd = False
                    # rotor assumes the space for input data but not input grad
                    for kdn_name in op.users_global:
                        if not input_grad and self.input_size[0] in kdn_name:
                            self.tmp[i:] += self.input_size[1]
                            input_grad = True

            self.save[i] += self.alive_list[i][:-2].dot(
                np.array(self.mem_sizes[:-2])
            )  # input kdn is not included
            # if (
            #     not output_grad
            #     and self.alive_list[i][
            #         self.kdn_names.index(self.output_size[0] + " grad")
            #     ]
            # ):
            #     self.save[i:] -= self.output_size[1]
            #     output_grad = True
        self.overhead = max(self.save + self.tmp) - self.save[-1]
        self.time = sum([op.time for op in self.op_list])

    def get_del_input_idx(self, kg):
        """
        This method is to find the idx where input is no longer needed.
        Should only used for Fn
        """
        input_kdn_name = kg.input_kdn_data.name
        for i, op in enumerate(self.op_list):
            if isinstance(op, RunOp) and input_kdn_name in op.deps_global:
                self.del_input_idx = i + 1

    def del_input(self):
        self.op_list.insert(self.del_input_idx, self.del_input_op)
        alive_status = self.alive_list[self.del_input_idx - 1].copy()
        self.alive_list.insert(self.del_input_idx, alive_status)
        for i in range(self.del_input_idx, len(self.op_list)):
            self.alive_list[i][-1] = False

        self.get_mem_time()
        # self.save = np.append(self.save, self.save[-1])
        # # self.del_input_idx = max(self.op_list.index(kcn)
        # #                     for kcn in input_kdn.users_global
        # #                     if kcn in self.op_list)
        # self.save[self.del_input_idx :] -= input_kdn.mem
        # self.overhead = max(self.save + self.tmp) - self.save[-1]

    def valid_sched(self):
        for i, op in enumerate(self.op_list[1:]):
            if hasattr(op, "deps_global"):
                for kdn_name in op.deps_global:
                    if (
                        not self.alive_list[i][self.kdn_names.index(kdn_name)]
                        and kdn_name not in op.deps_fake
                    ):
                        print(f"{kdn_name} is not alive when {op.name}")
                        return False
        return True


# class RK_Function:
#     def __init__(self, code_fe, code_fn, code_fc, code_bwd):
#         self.code_fe = code_fe
#         self.code_fn = code_fn
#         self.code_fc = code_fc
#         self.code_bwd = code_bwd

#     def exec_fe(self, storage: RK_Storage):
#         self.code_fe.exec(storage)

#     def exec_fn(self, storage: RK_Storage):
#         self.code_fn.exec(storage)

#     def exec_fc(self, storage: RK_Storage):
#         self.code_fc.exec(storage)

#     def exec_bwd(self, storage: RK_Storage):
#         self.code_bwd.exec(storage)



# ==========================
# definition file of RK_Sequence
# based on rotor/algorithms/sequence.py
# ==========================

from rockmate.def_op import OpSchedule

ref_print_atoms = [True]

# ==========================
#  ====== Seq Atom Op =======
# ==========================
class SeqAtomOp:
    def __init__(self, op):
        header = f"{op.op_type} {op.main_target}"
        self.name = header
        self.time = op.time
        self.op = op

    def __str__(self):
        return self.name


# ==========================


# ==========================
# ========= Seq Op =========
# ==========================
class SeqOp:
    pass


class SeqLoss(SeqOp):
    def __init__(self):
        self.time = 0
        self.mem = 0

    def __str__(self):
        return "Loss"


# ** Seq Block Op **
# -> Subclasses : Fn,Fc,Fe,Bwd
class SeqBlockOp(SeqOp):
    def __init__(self, name, index, op_sched: OpSchedule):
        self.name = name
        self.index = index
        self.op_sched = op_sched
        self.time = self.op_sched.time  # sum(o.time for o in body)
        self.mem = self.op_sched.save[-1]  # sum(o.mem  for o in body)
        self.overhead = self.op_sched.overhead

    def __str__(self):
        header = f"{self.name} Block {self.index} in {self.time}"
        if ref_print_atoms[0]:
            sb = "\n|  ".join([o.__str__() for o in self.op_sched.op_list])
            return f"{header}\n{sb}"
        else:
            return header


class SeqBlockFn(SeqBlockOp):
    def __init__(self, *args):
        super().__init__("Fn", *args)


class SeqBlockFc(SeqBlockOp):
    def __init__(self, *args):
        super().__init__("Fc", *args)


class SeqBlockFe(SeqBlockOp):
    def __init__(self, *args):
        super().__init__("Fe", *args)


class SeqBlockBwd(SeqBlockOp):
    def __init__(self, *args):
        super().__init__("Bwd", *args)


# ==========================


# ==========================
# ======== Sequence ========
# ==========================


class RK_Sequence:
    def __init__(self, l=None):
        if l:
            self.seq = l
        else:
            self.seq = []

    def __str__(self):
        return "\n".join([str(o) for o in self.seq])

    def insert(self, op: SeqOp):
        self.seq.append(op)

    def insert_seq(self, sub_seq):
        self.seq.extend(sub_seq.seq)

    def compute_time(self):
        return sum([o.time for o in self.seq])

    def cut_fwd_bwd(self):
        ln = len(self.seq)
        for i in range(ln):
            if isinstance(self.seq[i], SeqLoss):
                return (
                    RK_Sequence(list(self.seq[:i])),
                    RK_Sequence(list(self.seq[(i + 1) :])),
                )
        raise Exception("Can't cut a Sequence which doesn't have SeqLoss")


# ==========================



from rkgb.utils.ast_add_on import make_str_assign, make_str_list_assign
from rkgb.utils import np, torch
from rockmate.def_op import DelOp

# region Define Register Hooks
def fct_get_pack(storage, no_save_list, sanity_check=False):
    # no_save_list contains a list of names
    def pack(x):
        for i, c in enumerate(no_save_list):
            if storage.ld[c].data_ptr() == x.data_ptr():
                if sanity_check:
                    assert torch.equal(
                        storage.ld[c].data.as_strided_(
                            x.shape, x.stride(), x.storage_offset()
                        ),
                        x,
                    )
                return (
                    c,
                    x.shape,
                    x.stride(),
                    x.storage_offset(),
                    # x.clone(),
                )
        return x

    return pack


def fct_get_unpack(storage):
    def unpack(x):
        if isinstance(x, tuple):
            return storage.ld[x[0]].data.as_strided_(*x[1:4])

        return x

    return unpack

    # endregion


# region Basic Functions


def fct_get_shapes(storage, tensor_name):
    def fct():
        storage.shapes[tensor_name] = storage.ld[tensor_name].shape
        storage.dtypes[tensor_name] = storage.ld[tensor_name].dtype

    return fct


def fct_get_rng_state(storage, op_name):
    def fct():
        storage.rng_state.get(op_name)

    return fct


def fct_restore_rng_state(storage, op_name):
    def fct():
        storage.rng_state.restore(op_name)

    return fct


def fct_run_forward_no_grad(storage, code):
    def fct():
        with torch.no_grad():
            exec(code, storage.gd, storage.ld)

    return fct


def fct_run_forward_with_grad(storage, code, no_save_list=[]):
    def fct():
        with torch.autograd.graph.saved_tensors_hooks(
            fct_get_pack(storage, no_save_list), fct_get_unpack(storage)
        ):
            exec(code, storage.gd, storage.ld)

    return fct


def fct_run_inplace(storage, tensor_name, inplace_code):
    def fct():
        # ld = {tensor_name: storage.ld[f"_{tensor_name}"]}
        # code = f"{tensor_name} = x\n{inplace_code}"
        # print(tensor_name, inplace_code)
        exec(inplace_code, storage.gd, storage.ld)

    return fct


def fct_run_detach(storage, tensor_name):
    def fct():
        storage.ld[tensor_name].data = storage.ld[f"_{tensor_name}"].data

    return fct


def fct_assign_proxy(storage, tensor_name):
    def fct():
        storage.ld[f"_{tensor_name}"] = storage.ld[tensor_name]

    return fct


def fct_requires_grad(storage, tensor_name):
    def fct():
        storage.ld[tensor_name].requires_grad_()

    return fct


def fct_run_backward(storage, tensor_name, retain_graph):
    def fct():
        storage.ld[f"_{tensor_name}"].backward(
            storage.ld[tensor_name].grad, retain_graph=retain_graph
        )

    return fct


def fct_run_backward_with_inputs(
    storage, tensor_name, retain_graph, input_names
):
    inputs = [storage.ld[name] for name in input_names]

    def fct():
        storage.ld[f"_{tensor_name}"].backward(
            storage.ld[tensor_name].grad,
            inputs=inputs,
            retain_graph=retain_graph,
        )

    return fct


def fct_generate_fake_data(storage, tensor_name):
    def fct():
        m = (
            storage.gd["cmeta"]
            if storage.dtypes[tensor_name].is_complex
            else storage.gd["meta"]
        )
        x = m.expand(np.prod(storage.shapes[tensor_name]))
        storage.ld[tensor_name].data = x.view(storage.shapes[tensor_name])

    return fct


def fct_del_tensor_data(storage, tensor_name):
    def fct():
        storage.ld[tensor_name].data = torch.empty(
            0, device=storage.gd["device"]
        )

    return fct


def fct_del_tensor_base(storage, tensor_name):
    def fct():
        storage.ld[f"_{tensor_name}"]._base.data = torch.empty(
            0, device=storage.gd["device"]
        )

    return fct


def fct_del_tensor_grad(storage, tensor_name):
    def fct():
        storage.ld[tensor_name].grad = None

    return fct


def fct_del_var(storage, var_name):
    def fct():
        storage.ld[var_name] = None

    return fct

    # endregion


class RngState:
    def __init__(self):
        self.cpu_states = {}
        self.gpu_states = {}

    def get(self, op_name):
        if op_name not in self.cpu_states.keys():
            self.cpu_states[op_name] = torch.get_rng_state()
            self.gpu_states[op_name] = torch.cuda.get_rng_state()

    def restore(self, op_name):
        # pass
        torch.set_rng_state(self.cpu_states[op_name])
        torch.cuda.set_rng_state(self.gpu_states[op_name])


class RK_Storage:
    def __init__(self, device, nn_mod, dict_constants):
        self.gd = {
            **globals(),
            **dict_constants,
            "original_mod": nn_mod,
            "device": device,
            "torch": torch,
            "meta": torch.ones(1).to(device),
            "cmeta": torch.view_as_complex(torch.ones(2)).to(device),
        }
        self.ld = {}
        self.shapes = dict()
        self.dtypes = dict()
        self.rng_state = RngState()

    def add_val(self, val, x):
        self.ld[val] = x

    def get_val(self, val):
        try:
            return self.ld[val]
        except:
            try:
                return self.gd[val]
            except:
                raise Exception(f"{val} not in the storage")


class Compiler:
    """
    The compiler takes the full operation schedule as input,
    return the lists of Python functions.
    Each list corresponds to one operation.
    """

    def __init__(self, storage):
        self.storage = storage
        self.shapes = storage.shapes
        self.device = self.storage.gd["device"]

    def _is_alive(self, kdn_name, i):
        if kdn_name in self.op_sched.kdn_names:
            return self.op_sched.alive_list[i][
                self.op_sched.kdn_names.index(kdn_name)
            ]

        else:
            return True

    def find_next_idx(l, target, i):
        return i + l[i:].index(target)

    def get_fwd(self, op, i):
        if "loss" in op.main_target:
            return [fct_run_forward_no_grad(self.storage, "")]
        rec = op.name in self.op_sched.op_name_list[:i]
        if not op.proxy:
            last_before_bwd = False
        else:
            next_bwd_idx = i + self.op_sched.op_name_list[i:].index(
                op.name.replace("fwd", "bwd")
            )
            last_before_bwd = not (
                op.name in self.op_sched.op_name_list[i + 1 : next_bwd_idx]
            )
        l = []

        if op.is_rand:
            if not rec:
                l.append(fct_get_rng_state(self.storage, op.name))
            else:
                l.append(fct_restore_rng_state(self.storage, op.name))

        # compile inplace code
        inplace_code = make_str_list_assign(
            op.inplace_code, force_special_kwargs=rec
        )
        # compile body code
        body_code = ""
        for bc in op.body_code:
            suffix = ""
            if rec and (bc[0] in op.tensor_targets):
                suffix = ".data"
            body_code += (
                make_str_assign(bc, suffix=suffix, force_special_kwargs=rec)
                + "\n"
            )

        # compile main code
        suffix = ""
        main_code = (
            make_str_assign(
                op.main_code, suffix=suffix, force_special_kwargs=rec
            )
            + "\n"
        )
        main_code = main_code.replace(op.main_target, f"_{op.main_target}")

        if not last_before_bwd:

            # inplace_code = inplace_code.replace(
            #     op.main_target, f"_{op.main_target}"
            # )

            for target in op.tensor_targets:
                inplace_code = inplace_code.replace(target, "_" + target)
            l.append(
                fct_run_forward_no_grad(
                    self.storage, main_code.replace("self", "original_mod"),
                )
            )
        else:
            no_save_list = []
            candidates = list(op.deps_global) + list(op.users_global)
            for kdn_name in candidates:
                if kdn_name in self.op_sched.op_name_list[i:next_bwd_idx]:
                    no_save_list.append(kdn_name.split(" ")[0])

            for target in op.tensor_targets:
                inplace_code = inplace_code.replace(target, "_" + target)

            l.append(
                fct_run_forward_with_grad(
                    self.storage,
                    main_code.replace("self", "original_mod"),
                    no_save_list=no_save_list,
                )
            )
        l.append(
            fct_run_forward_with_grad(
                self.storage, inplace_code.replace("self", "original_mod"),
            )
        )
        l.append(fct_run_detach(self.storage, op.main_target))
        l.append(
            fct_run_forward_with_grad(
                self.storage, body_code.replace("self", "original_mod")
            )
        )

        # get the shape of tensors
        if not rec:
            l.append(fct_get_shapes(self.storage, f"_{op.main_target}"))
            for target in op.tensor_targets:
                l.append(fct_get_shapes(self.storage, target))
        return l

    def get_bwd(self, op, i):
        rec = op.name in self.op_sched.op_name_list[:i]
        last = not (op.name in self.op_sched.op_name_list[i + 1 :])
        l = []
        l2 = []

        if op.is_rand:
            if not rec:
                l.append(fct_get_rng_state(self.storage, op.name))
            else:
                l.append(fct_restore_rng_state(self.storage, op.name))

        temporary_tensor_names = [
            kdn_name.split(" ")[0]
            for kdn_name in op.deps_fake
            if not self._is_alive(kdn_name, i)
        ]
        if op.main_target in temporary_tensor_names:
            temporary_tensor_names.append(f"_{op.main_target}")
        for tensor_name in temporary_tensor_names:
            l.append(fct_generate_fake_data(self.storage, tensor_name))
            l2.append(fct_del_tensor_data(self.storage, tensor_name))
        if rec:
            prev_i = i - self.op_sched.op_name_list[:i][::-1].index(op.name) - 1
            input_names = []
            for kdn_name in op.users_global:
                if f"del {kdn_name}" in self.op_sched.op_name_list[prev_i:i]:
                    input_names.append(kdn_name.split(" ")[0])
            l.append(
                fct_run_backward_with_inputs(
                    self.storage,
                    op.main_target,
                    retain_graph=(not last),
                    input_names=input_names,
                )
            )
        else:
            l.append(
                fct_run_backward(
                    self.storage, op.main_target, retain_graph=(not last)
                )
            )

        return l + l2

    def get_del_data(self, op, i):
        l = []
        l.append(fct_del_tensor_data(self.storage, op.main_target))
        if op.info is not None and op.info.requires_grad:
            l.append(fct_del_tensor_data(self.storage, f"_{op.main_target}"))
        if op.includes_base:
            l.append(fct_del_tensor_base(self.storage, op.main_target))
        for v in op.tensor_targets:
            l.append(fct_del_tensor_data(self.storage, v))
        for v in op.container_targets:
            l.append(fct_del_var(self.storage, v))
        # l.append(fct_del_var(self.storage, f"_{op.main_target}"))

        return l

    def get_del_grad(self, op, i):
        return [fct_del_tensor_grad(self.storage, op.main_target)]

    def compile(self, op_sched):
        self.op_sched = op_sched

        fct_list = []
        for i, op in enumerate(op_sched.op_list):
            if "fwd" in op.name:
                fct_list.append(self.get_fwd(op, i))
            elif "bwd" in op.name:
                fct_list.append(self.get_bwd(op, i))
            elif "data" in op.name:
                fct_list.append(self.get_del_data(op, i))
            elif "grad" in op.name:
                fct_list.append(self.get_del_grad(op, i))
            else:
                fct_list.append([])

        return fct_list




from setuptools import setup, Extension

setup(
    ext_modules=[
        Extension(
            "rockmate.csolver",
            ["src/solver.c"],
            extra_compile_args=["-O3", "--std=c99"],
        ),
    ],
)




# ==========================
# definition file of RK_Chain
# also contains RK_Chain builder -> depends on use_chk.py
#  based on rotor/algorithms/parameters.py
# ==========================
from rkgb.utils import imports_from_rotor as irotor
from rockmate.ILP_gurobi_solver import ModelGurobi
import numpy as np
from rockmate.def_op import RunOp, DelOp, OpSchedule
import math

# ==========================
# ======== RK Block ========
# ==========================


def get_rk_solution(list_kg, l_bd_abar, budget_all):

    if True: ##only support for Gurobi solver for now
        param_dict = {
            "LogToConsole": 0,
            "IntegralityFocus": 1,
        }
        md = ModelGurobi(
            list_kg[0],
            budget_all,
            max(l_bd_abar),
            gcd=10000,
            gurobi_params=param_dict,
        )

    else:
        md = ModelMIP(list_kg[0], budget_all, max(l_bd_abar), gcd=10000,)
        md.md.verbose = 0
    list_list_sols = []
    for bd_abar in np.sort(l_bd_abar)[::-1]:
        md.add_abar_constraint(bd_abar)
        md.solve()

        if not md.feasible:
            list_list_sols.append(False)
            continue
            # return False
        list_sols = []
        for kg in list_kg:
            fwd_sched, bwd_sched = md.schedule(kg)
            list_sols.append(RK_Block_Solution(fwd_sched, bwd_sched))
        list_list_sols.append(list_sols)
    return list_list_sols


class RK_Block_Solution:
    def __init__(self, fwd_sched, bwd_sched):
        self.fwd_sched, self.bwd_sched = fwd_sched, bwd_sched
        self.time_fwd = self.fwd_sched.time
        self.time_bwd = self.bwd_sched.time
        self.size_a_bar = self.fwd_sched.save[-1]
        self.overhead_fwd = self.fwd_sched.overhead
        self.overhead_bwd = (
            self.bwd_sched.overhead + self.bwd_sched.save[-1] - self.size_a_bar
        )


def get_rk_block(list_kg, nb_bdg_abar, nb_bdg_all):
    list_blocks = []
    for kg in list_kg:
        list_blocks.append(RK_Block(kg))
    kdn_sizes = [kdn.mem for kdn in kg.list_kdn]
    overheads = [kcn.overhead for kcn in kg.list_kcn]
    max_bdg = sum(kdn_sizes) + max(overheads)
    min_bdg = (
        list_blocks[-1].Fc_sched.overhead + list_blocks[-1].Fc_sched.save[-1]
    )
    l_bd_all = np.linspace(min_bdg, max_bdg, nb_bdg_all)
    sols = []
    uniq_sols = set()
    for bd_all in l_bd_all:
        l_bd_abar = np.linspace(kg.output_kdn_data.mem, bd_all, nb_bdg_abar)
        # for bd_abar in l_bd_abar:
        #     if bd_all >= bd_abar:
        list_sols = get_rk_solution(list_kg, l_bd_abar, bd_all)
        for sol in list_sols:
            if sol:
                t = (
                    sol[0].size_a_bar,
                    sol[0].overhead_fwd,
                    sol[0].overhead_bwd,
                )
                if not (t in uniq_sols):
                    uniq_sols.add(t)
                    sols.append(sol)
                    for s, block in zip(sol, list_blocks):
                        block.sols.append(s)
    return list_blocks


class RK_Block:
    def __init__(self, kg):
        self.block_name = (
            f"Block[{kg.input_kdn_data.name}->{kg.output_kdn_data.name}]"
        )
        self.sols = []
        # == build Fc/Fn schedule
        def _fast_fwd_sched():
            def _can_del(i, kdn):
                for kcn in kdn.users_real:
                    if "bwd" in kcn.name:
                        continue
                    if kg.list_kcn.index(kcn) > i:
                        return False
                return True

            op_list = []
            alive_list = []
            alive_status = np.zeros(len(kg.list_kdn) + 2, dtype=bool)
            alive_status[-1] = True
            loss_idx = kg.list_kcn.index(kg.loss_kcn)
            for i, kcn in enumerate(kg.list_kcn[:loss_idx]):
                op = RunOp(kcn)
                op.no_grad = True
                op_list.append(op)
                for kdn in kcn.users:
                    if "data" not in kdn.kdn_type:
                        continue
                    alive_status[kg.list_kdn.index(kdn)] = 1
                alive_list.append(alive_status.copy())
                for j, kdn in enumerate(kg.list_kdn):
                    if kdn in [kg.output_kdn_data, kg.output_kdn_grad]:
                        continue
                    if alive_status[j] and _can_del(i, kdn):
                        op = DelOp(kdn)
                        op.proxy = False
                        op_list.append(op)
                        alive_status[j] = 0
                        alive_list.append(alive_status.copy())
            return op_list, alive_list

        self.Fc_sched = OpSchedule(
            *_fast_fwd_sched(),
            kg.input_kdn_data,
            kg.input_kdn_grad,
            kg.output_kdn_data,
            kg.list_kdn,
            no_grad=True,
        )
        self.Fn_sched = OpSchedule(
            *_fast_fwd_sched(),
            kg.input_kdn_data,
            kg.input_kdn_grad,
            kg.output_kdn_data,
            kg.list_kdn,
            no_grad=True,
        )
        self.Fn_sched.get_del_input_idx(kg)
        self.Fn_sched.del_input()
        self.overhead_fast_fwd = self.Fc_sched.overhead
        self.time_fast_fwd = self.Fc_sched.time

        #  == build .mem_inp/out ==
        # memsize = lambda inp : kg.dict_info[inp].memsize
        self.mem_inp = kg.input_kdn_data.mem if kg.input_kdn_data.mem else 0
        self.mem_out = kg.output_kdn_data.mem if kg.output_kdn_data.mem else 0

    def __str__(self):
        return (
            f"\n~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~\n"
            f"{self.block_name} :\n"
            f"\tnb of sol : {len(self.sols)}\n"
            f"\tmem_inp   : {irotor.MemSize(self.mem_inp)}\n"
            f"\ttime_ff  : {self.time_ff}\n"
            f"\n~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~\n"
        )


# ==========================
# ======== RK CHAIN ========
# ==========================


class RK_Chain:
    def __init__(
        self,
        list_kg,
        eq_classes,
        nb_budget_abar=10,
        nb_budget_all=3,
        mem_unit=None,
    ):
        if mem_unit:
            self.mem_unit = mem_unit
        else:
            self.mem_unit = 1024 ** 2
        self.body = [None] * len(list_kg)
        for cls in eq_classes:
            l_kg = [list_kg[i] for i in cls]
            l_block = get_rk_block(l_kg, nb_budget_abar, nb_budget_all)
            for i, j in enumerate(cls):
                self.body[j] = l_block[i]

        # for l_kg in identical_kg:
        #     self.body += get_rk_block(l_kg, nb_budget_abar, nb_budget_all)

        # for g in list_kg:
        #     self.body.append(RK_Block(g,nb_budget_abar,nb_budget_all))
        #     print_debug(self.body[-1])
        # organizes the information for rotor_solver.py as in Rotor
        # -> fw/bw/cw/cbw/fwd_tmp/bwd_tmp
        # -> in those list, one dummy block is added at the end for Loss
        # fw/bw: runtime of fwd/bwd
        # cbw: saved memory in each solution
        # cw: saved memory for each checkpoint solution (only the input)

        # -- init variables --
        ln = len(self.body)
        mkl = lambda n: [[] for _ in range(n)]
        fw = mkl(ln + 1)
        bw = mkl(ln + 1)
        cw = [None] * (ln + 2)
        cbw = mkl(ln + 2)
        fwd_tmp = mkl(ln + 1)
        bwd_tmp = mkl(ln + 1)
        ff_fwd_tmp = [None] * (ln + 1)
        ff_fw = [None] * (ln + 1)
        nb_sol = []

        # -- extract info from each block
        for (i, b) in enumerate(self.body):
            nb_sol.append(len(b.sols))
            if nb_sol[-1] == 0:
                raise Exception(
                    f"We need at least one solution per block. "
                    f"Here {b.block_name} has no solution"
                )
            for sol in b.sols:
                fw[i].append(sol.time_fwd)
                bw[i].append(sol.time_bwd)
                cbw[i + 1].append(sol.size_a_bar)
                fwd_tmp[i].append(sol.overhead_fwd)
                bwd_tmp[i].append(sol.overhead_bwd)
            cw[i] = b.mem_inp
            ff_fwd_tmp[i] = b.overhead_fast_fwd
            ff_fw[i] = b.time_fast_fwd
        cw[ln] = self.body[-1].mem_out  #  the final output

        # for the Loss block :
        nb_sol.append(1)
        fw[-1] = [0]
        bw[-1] = [0]
        cw[-1] = 0
        cbw[-1] = [0]
        fwd_tmp[-1] = [0]
        bwd_tmp[-1] = [0]
        ff_fwd_tmp[-1] = 0
        ff_fw[-1] = 0

        #  return :
        self.ln = ln
        self.fw = fw
        self.bw = bw
        self.cw = self.discretize(cw)
        self.cbw = [self.discretize(x) for x in cbw]
        self.fwd_tmp = [self.discretize(x) for x in fwd_tmp]
        self.bwd_tmp = [self.discretize(x) for x in bwd_tmp]
        self.ff_fwd_tmp = self.discretize(ff_fwd_tmp)
        self.ff_fw = ff_fw
        self.nb_sol = nb_sol

    def discretize(self, values):
        return [math.ceil(value / self.mem_unit) for value in values]


# ==========================
